import 'dart:convert';

import 'package:flutter/material.dart';
import 'package:flutter_secured_storage/flutter_secured_storage.dart';
import 'package:m4e_sdk/m4e_sdk.dart';

import 'package:sdk_example/services/wallet_service.dart';

enum WalletRepositoryState { initial, loading, success, error }

class WalletBalance {
  final M4eWallet wallet;
  final M4eBalance balance;

  const WalletBalance({@required this.wallet, @required this.balance});

  @override
  bool operator ==(Object o) {
    if (identical(this, o)) return true;

    return o is WalletBalance && o.wallet == wallet && o.balance == balance;
  }

  @override
  int get hashCode => wallet.hashCode ^ balance.hashCode;

  @override
  String toString() => 'WalletBalance(wallet: $wallet, balance: $balance)';
}

class WalletRepository with ChangeNotifier {
  final _walletService = WalletService();
  final _storage = FlutterSecuredStorage();

  static M4eWalletList _wallets;
  M4eWalletList get wallets => _wallets;

  static List<WalletBalance> _walletBalances = [];
  List<WalletBalance> get walletBalances => _walletBalances;

  static M4eException _error;
  M4eException get error => _error;

  static WalletRepositoryState _state;
  WalletRepositoryState get state => _state ?? WalletRepositoryState.initial;

  static WalletRepositoryState _balanceState;
  WalletRepositoryState get balanceState =>
      _state ?? WalletRepositoryState.initial;

  Future<void> getWalletBallance() async {
    try {
      _balanceState = WalletRepositoryState.loading;
      notifyListeners();

      // get wallets-balances for each wallet
      wallets.wallets.forEach((wallet) async {
        final _walletBalance = await _walletService.getWalletBalance(wallet.id);

        final _wallet = wallets.wallets
            .firstWhere((w) => w.id == wallet.id, orElse: () => null);

        // create wallet balance
        final _walletWithBalance =
            WalletBalance(wallet: _wallet, balance: _walletBalance);

        // add to wallet balance list
        final _walletHasBalance = _walletBalances?.firstWhere(
            (walletBalance) =>
                walletBalance.wallet.id == _walletWithBalance.wallet.id,
            orElse: () => null);

        // map balance to wallet
        if (_walletHasBalance == null) {
          // add to wallet-balance to liist
          _walletBalances.add(_walletWithBalance);
          _balanceState = WalletRepositoryState.success;
          notifyListeners();
        } else {
          // get index
          final _walletIndex = _walletBalances.indexWhere((walletBalance) =>
              walletBalance.wallet.id == _walletWithBalance.wallet.id);

          // update existing
          _walletBalances[_walletIndex] = _walletWithBalance;
          _balanceState = WalletRepositoryState.success;
          notifyListeners();
        }
      });
    } on M4eException catch (e) {
      _error = e;

      _balanceState = WalletRepositoryState.error;
      notifyListeners();
    }
  }

  Future<void> getWallets() async {
    try {
      _state = WalletRepositoryState.loading;
      notifyListeners();

      // get user from cache
      final _userString = await _storage.read(key: 'M4E_USER');

      if (_userString != null) {
        final _m4eUser = M4eUser.fromJson(json.decode(_userString));

        // get wallets with user-id
        _wallets = await _walletService.getWallets(_m4eUser.id);

        _state = WalletRepositoryState.success;
        notifyListeners();

        await getWalletBallance();
      } else {
        _error = M4eExceptionMessages.kUnauthorized;

        _state = WalletRepositoryState.error;
        notifyListeners();
      }
    } on M4eException catch (e) {
      _error = e;

      _state = WalletRepositoryState.error;
      notifyListeners();
    }
  }

  Future<void> createWallet({@required M4eWalletForm walletForm}) async {
    try {
      _state = WalletRepositoryState.loading;
      notifyListeners();

      // get user from cache
      final _userString = await _storage.read(key: 'M4E_USER');

      if (_userString != null) {
        final _m4eUser = M4eUser.fromJson(json.decode(_userString));

        // get wallets with user-id
        await _walletService.createWallet(_m4eUser.id, walletForm);

        _state = WalletRepositoryState.success;
        notifyListeners();
      } else {
        _error = M4eExceptionMessages.kUnauthorized;

        _state = WalletRepositoryState.error;
        notifyListeners();
      }
    } on M4eException catch (e) {
      _error = e;

      _state = WalletRepositoryState.error;
      notifyListeners();
    }
  }
}
