import 'dart:convert';

import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:fluttertoast/fluttertoast.dart';
import 'package:m4e_sdk/m4e_sdk.dart';
import 'package:provider/provider.dart';

import 'package:sdk_example/repositories/transaction_repository.dart';
import 'package:sdk_example/screens/transaction_screens/widgets/wallet_drop_down.dart';

class TransactionMeta {
  final String note;
  final String callbackUrl;

  const TransactionMeta({@required this.note, @required this.callbackUrl});

  Map<String, dynamic> toMap() {
    return {
      'note': note,
      'callbackUrl': callbackUrl,
    };
  }

  factory TransactionMeta.fromMap(Map<String, dynamic> map) {
    if (map == null) return null;

    return TransactionMeta(
      note: map['note'],
      callbackUrl: map['callbackUrl'],
    );
  }

  String toJson() => json.encode(toMap());

  factory TransactionMeta.fromJson(String source) =>
      TransactionMeta.fromMap(json.decode(source));
}

// TODO: add m4e reciepient-contact in transaction form (to make transactions with user contacts)
class TransactionFormScreen extends StatefulWidget {
  @override
  _TransactionFormScreenState createState() => _TransactionFormScreenState();
}

class _TransactionFormScreenState extends State<TransactionFormScreen> {
  final _formKey = GlobalKey<FormState>();

  M4eWallet _selectedSourceWallet;
  M4eWallet _selectedReciepienteWallet;

  final _amountController = TextEditingController();
  final _reciepientPhoneNumberController = TextEditingController();
  final _referenceController = TextEditingController();

  @override
  void initState() {
    super.initState();

    // Future.microtask(() => _createNewTransaction());
  }

  void _createNewTransaction() {
    final _repository =
        Provider.of<TransactionRepository>(context, listen: false);

    final _phoneNumber = M4ePhoneNumber(
      country: 'GHANA',
      number: _reciepientPhoneNumberController.text,
    );

    final _transactionMeta = TransactionMeta(
      callbackUrl: 'www.my-custom-callbackUrl.com',
      note: _referenceController.text,
    );

    final _transactionform = M4eTransactionForm(
      // TODO: properly implement transaction wallet case
      from: _selectedSourceWallet == null
          ? M4eTransactionWallet.mobileMoneyWalletProvider(
              provider: M4eNetworkProviderType.MTN,
              phoneNumber: _phoneNumber,
            )
          : M4eTransactionWallet.m4eWalletProvider(
              wallet: _selectedSourceWallet,
            ),
      to: _selectedReciepienteWallet == null
          ? M4eTransactionWallet.mobileMoneyWalletProvider(
              provider: M4eNetworkProviderType.MTN,
              phoneNumber: _phoneNumber,
            )
          : M4eTransactionWallet.m4eWalletProvider(
              wallet: _selectedReciepienteWallet,
            ),
      amount: M4eAmount(
          currency: M4eCurrencyType.GHS,
          value: double.parse(_amountController.text)),
      at: DateTime.now(),
      meta: _transactionMeta.toMap(),
    );

    if (!_formKey.currentState.validate()) return;

    _repository.createTransaction(transactionForm: _transactionform).then((_) {
      if (_repository.state == TransactionRepositoryState.error) {
        Fluttertoast.showToast(
            msg: '${_repository.error.code}\n${_repository.error.message}');

        print(_repository.error.message);
      } else {
        // _formKey.currentState.reset();

        Fluttertoast.showToast(msg: 'Wallet created');
      }
    });
  }

  Widget _buildAmountTextField() {
    return TextFormField(
      controller: _amountController,
      keyboardType: TextInputType.numberWithOptions(),
      textAlignVertical: TextAlignVertical.center,
      decoration: InputDecoration(
        prefixText: 'GHS ',
        hintText: 'Enter amount',
      ),
      validator: (_) {
        return _amountController.text.isEmpty ? 'Enter amount' : null;
      },
    );
  }

  Widget _buildReceiverPhoneNumberTextField() {
    return TextFormField(
      keyboardType: TextInputType.phone,
      controller: _reciepientPhoneNumberController,
      textAlignVertical: TextAlignVertical.center,
      decoration: InputDecoration(
        prefixIcon: Icon(Icons.phone),
        hintText: 'Enter receiver phone number',
      ),
      validator: (_) {
        return _reciepientPhoneNumberController.text.isEmpty
            ? 'Enter receiver phone number'
            : null;
      },
    );
  }

  Widget _buildReferenceTextField() {
    return TextFormField(
      maxLines: 3,
      keyboardType: TextInputType.multiline,
      controller: _referenceController,
      textAlignVertical: TextAlignVertical.center,
      decoration: InputDecoration(
        hintText: 'Enter reference',
      ),
      validator: (_) {
        return _referenceController.text.isEmpty ? 'Enter reference' : null;
      },
    );
  }

  Widget _buildSubmitFormFloatingActionButton() {
    return Consumer<TransactionRepository>(
      builder: (context, repository, __) {
        print(repository.state);

        return FloatingActionButton(
          onPressed: () => _createNewTransaction(),
          child: repository.state == TransactionRepositoryState.loading
              ? CircularProgressIndicator(
                  backgroundColor: Theme.of(context).primaryColor)
              : Icon(Icons.check),
        );
      },
    );
  }

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      appBar: AppBar(
        title: Text('Create transaction'),
      ),
      body: Form(
        key: _formKey,
        child: ListView(
          padding: EdgeInsets.only(left: 30.0, right: 30.0, top: 30.0),
          children: [
            WalletDropDownTextField(
              hintText: 'Select source wallet',
              selectedWallet: _selectedSourceWallet,
              onSelectWalletCallback: (wallet) {
                setState(() {
                  _selectedSourceWallet = wallet;
                });
                print(_selectedSourceWallet);
              },
            ),
            SizedBox(height: 20.0),
            _buildAmountTextField(),
            SizedBox(height: 20.0),
            _buildReceiverPhoneNumberTextField(),
            SizedBox(height: 20.0),
            _buildReferenceTextField(),
            SizedBox(height: 30.0),
            _buildSubmitFormFloatingActionButton(),
            SizedBox(height: 20.0),
          ],
        ),
      ),
    );
  }
}
