import 'package:flutter/material.dart';
import 'package:fluttertoast/fluttertoast.dart';
import 'package:provider/provider.dart';
import 'package:m4e_sdk/m4e_sdk.dart';
import 'package:sdk_example/repositories/wallet_repository.dart';

class WalletDropDownTextField extends StatefulWidget {
  final String hintText;
  final M4eWallet selectedWallet;
  final Function(M4eWallet) onSelectWalletCallback;

  const WalletDropDownTextField({
    Key key,
    @required this.hintText,
    @required this.selectedWallet,
    @required this.onSelectWalletCallback,
  }) : super(key: key);

  @override
  _WalletDropDownTextFieldState createState() =>
      _WalletDropDownTextFieldState();
}

class _WalletDropDownTextFieldState extends State<WalletDropDownTextField> {
  // for validation only
  M4eWallet _selectedWallet;

  @override
  void initState() {
    super.initState();
    Future.microtask(() => _getUserWallets());
  }

  void _getUserWallets() {
    final _repository = Provider.of<WalletRepository>(context, listen: false);

    _repository.getWallets().then((_) {
      if (_repository.state == WalletRepositoryState.error) {
        Fluttertoast.showToast(
            msg: '${_repository.error.code}\n${_repository.error.message}');
      }
    });
  }

  List<DropdownMenuItem<M4eWallet>> _buildLoadingM4eWalletListState() {
    return [
      DropdownMenuItem(
        child: Center(
          child: FittedBox(
            child: CircularProgressIndicator(
              strokeWidth: 2.0,
            ),
          ),
        ),
      ),
    ];
  }

  List<DropdownMenuItem<M4eWallet>> _buildErrorWalletListState(
      BuildContext context, M4eException exception) {
    return [
      DropdownMenuItem(
        onTap: () => _getUserWallets(),
        child: Center(
          child: Text.rich(
            TextSpan(
              text: '${exception.message} ',
              children: <TextSpan>[
                TextSpan(
                  text: 'Tap to Retry',
                  style: Theme.of(context).inputDecorationTheme.labelStyle,
                ),
              ],
            ),
          ),
        ),
      ),
    ];
  }

  List<DropdownMenuItem<M4eWallet>> _buildLoadedWalletList(
      M4eWalletList m4eWalletList) {
    return m4eWalletList.wallets
        .map(
          (wallet) => DropdownMenuItem(
            onTap: () {},
            value: wallet,
            child: Text.rich(
              TextSpan(
                text: '${wallet.walletInfo.label} ',
                children: <TextSpan>[
                  TextSpan(
                    text: '(${wallet.walletInfo.type.toString()})'
                        .toLowerCase()
                        .replaceAll('_', ' '),
                    style: TextStyle(color: Colors.grey),
                  )
                ],
              ),
            ),
          ),
        )
        .toList();
  }

  List<DropdownMenuItem<M4eWallet>> _buildDropDownItems(
      BuildContext context, WalletRepository repository) {
    if (repository.state == WalletRepositoryState.initial ||
        repository.state == WalletRepositoryState.loading) {
      return _buildLoadingM4eWalletListState();
    } else if (repository.state == WalletRepositoryState.error) {
      return _buildErrorWalletListState(context, repository.error);
    } else {
      return _buildLoadedWalletList(repository.wallets);
    }
  }

  @override
  Widget build(BuildContext context) {
    return Consumer<WalletRepository>(
      builder: (_, repository, __) {
        return DropdownButtonFormField<M4eWallet>(
          hint: Text(widget.hintText,
              style: Theme.of(context).inputDecorationTheme.labelStyle),
          decoration: InputDecoration(
            prefixIcon: Icon(Icons.account_balance_wallet),
          ),
          items: _buildDropDownItems(context, repository),
          value: widget.selectedWallet,
          onChanged: (value) {
            setState(() {
              _selectedWallet = value;
            });
            widget.onSelectWalletCallback(value);
          },
          validator: (_) {
            return _selectedWallet == null
                ? 'Select wallet for transaction!'
                : null;
          },
        );
      },
    );
  }
}
