import 'package:flutter/material.dart';
import 'package:fluttertoast/fluttertoast.dart';
import 'package:provider/provider.dart';
import 'package:sdk_example/repositories/wallet_repository.dart';
import 'package:sdk_example/widgets/wallet_detail_form.dart';
import 'package:m4e_sdk/m4e_sdk.dart';

class WalletFormScreen extends StatefulWidget {
  @override
  _WalletFormScreenState createState() => _WalletFormScreenState();
}

class _WalletFormScreenState extends State<WalletFormScreen> {
  final _walletlabelController = TextEditingController();

  final _currencyController = TextEditingController(text: 'GHS');
  final _phoneNumberController = TextEditingController();

  String _selectedWalletType = 'M4E';
  String _selectedNetworkProvider;

  final _formKey = GlobalKey<FormState>();

  void _createNewWallet() {
    final _repository = Provider.of<WalletRepository>(context, listen: false);

    final _walletId = M4eUniqueId();
    final _walletType = M4eWalletType.parse(_selectedWalletType);

    final _mobileMoneyWalletDetail =
        M4eWalletDetailType.mobileMoneyWalletDetail(
      provider: M4eNetworkProviderType.parse(_selectedNetworkProvider),
      phoneNumber: M4ePhoneNumber(
        country: 'GHANA',
        number: _phoneNumberController.text,
      ),
    );

    final _walletForm = M4eWalletForm(
      id: _walletId,
      label: _walletlabelController.text,
      currency: M4eCurrencyType.parse(_currencyController.text),
      type: M4eWalletType.parse(_selectedWalletType),
      detail: _walletType == M4eWalletType.M4E
          ? M4eWalletDetailType.m4eWalletDetail(id: _walletId)
          : _mobileMoneyWalletDetail,
    );

    if (!_formKey.currentState.validate()) return;

    _repository.createWallet(walletForm: _walletForm).then((_) {
      if (_repository.state == WalletRepositoryState.error) {
        Fluttertoast.showToast(
            msg: '${_repository.error.code}\n${_repository.error.message}');
      } else {
        _formKey.currentState.reset();

        Fluttertoast.showToast(msg: 'Wallet created');
      }
    });
  }

  Widget _buildLabelTextField() {
    return TextFormField(
      controller: _walletlabelController,
      textAlignVertical: TextAlignVertical.center,
      decoration: InputDecoration(
        prefixIcon: Icon(Icons.phone),
        hintText: 'Enter wallet name',
      ),
      validator: (_) {
        return _walletlabelController.text.isEmpty
            ? 'Enter walet name/title'
            : null;
      },
    );
  }

  Widget _buildCurrencyTextField() {
    return TextFormField(
      controller: _currencyController,
      readOnly: true,
      textAlignVertical: TextAlignVertical.center,
      validator: (_) {
        return _currencyController.text.isEmpty ? 'Enter currency' : null;
      },
    );
  }

  Widget _buildTypeTextField() {
    return DropdownButtonFormField(
      value: _selectedWalletType,
      items: <DropdownMenuItem>[
        DropdownMenuItem(child: Text('M4E wallet'), value: 'M4E'),
        DropdownMenuItem(
            child: Text('Mobile Money wallet'), value: 'MOBILE_MONEY'),
        DropdownMenuItem(child: Text('Bank wallet'), value: 'BANK'),
      ],
      onChanged: (value) {
        setState(() {
          _selectedWalletType = value;
        });
      },
      validator: (value) {
        return _selectedWalletType == null || _selectedWalletType.isEmpty
            ? 'Select wallet type'
            : null;
      },
    );
  }

  Widget _buildWalletDetialForm() {
    if (_selectedWalletType != 'M4E') {
      return WalletDetailForm(
        sectionLabel: 'Detail Section',
        type: WalletDetailFormType.mobileMoneyWalletDetail(
          selectedNetworkProvider: _selectedNetworkProvider,
          onSelectNetworkProvider: (provider) {
            setState(() {
              _selectedNetworkProvider = provider;
            });
          },
          phoneNumberController: _phoneNumberController,
        ),
      );
    }

    return Container();
  }

  Widget _buildSubmitFormFloatingActionButton() {
    return Consumer<WalletRepository>(
      builder: (context, repository, __) {
        print(repository.state);

        return FloatingActionButton(
          onPressed: () => _createNewWallet(),
          child: repository.state == WalletRepositoryState.loading
              ? CircularProgressIndicator(
                  backgroundColor: Theme.of(context).primaryColor)
              : Icon(Icons.check),
        );
      },
    );
  }

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      appBar: AppBar(
        title: Text('Add Wallet'),
      ),
      body: Form(
        key: _formKey,
        child: ListView(
          padding: EdgeInsets.only(left: 30.0, right: 30.0, top: 30.0),
          children: [
            _buildLabelTextField(),
            SizedBox(height: 20.0),
            _buildCurrencyTextField(),
            SizedBox(height: 20.0),
            _buildTypeTextField(),
            SizedBox(height: 20.0),
            _buildWalletDetialForm(),
            SizedBox(height: 30.0),
            _buildSubmitFormFloatingActionButton(),
            SizedBox(height: 20.0),
          ],
        ),
      ),
    );
  }
}
