import 'dart:convert';
import 'dart:io';

import 'package:dio/dio.dart';
import 'package:flutter_secure_storage/flutter_secure_storage.dart' as storage;
import 'package:flutter/foundation.dart';

import '../../core/core.dart';
import '../../models/models.dart';
import '../../m4e_forms/m4e_forms.dart';

/// Accessing the instantiated `UserApi` methods
/// is done by passing its type to the GetIt instance
///
/// Example use of `UserApi`:
/// ```
/// import 'package:m4edart/m4e_sdk_initialize.dart' as sdk;
///
/// sdk.m4e<UserApi>().registerUser(M4eUserForm userForm);
/// ```
///
abstract class UserApi {
  /// Returns `M4eUser` with `M4ePhoneNumber`
  ///
  /// Throws `AuthException` with code `NO_AUTH_USER`
  /// when no user is authenticated
  ///
  /// Throws `ServerException` if call to server fails
  ///
  /// Throws `ServerException` with `NO_INTERNET_CONNECTION` error
  /// code device is not connected to the internet
  ///
  Future<M4eUser> getM4eUser(M4ePhoneNumber phoneNumber);

  /// Creates an m4e user account
  ///
  /// Throws `AuthException` with code `NO_AUTH_USER`
  /// when no user is authenticated
  ///
  /// Throws `ServerException` if call to server fails
  ///
  /// Throws `ServerException` with `NO_INTERNET_CONNECTION` error
  /// code device is not connected to the internet
  ///
  Future<M4eUniqueId> createM4eUser(M4eUniqueId userId, M4eUserForm userForm);
}

class UserApiImpl implements UserApi {
  final Dio _client;

  UserApiImpl({
    @required Dio client,
    @required storage.FlutterSecureStorage storage,
    @required M4eLoggingInterceptors loggingInterceptor,
    @required M4eHeaderInterceptors headerInterceptor,
  })  : assert(client != null),
        assert(storage != null),
        assert(loggingInterceptor != null),
        assert(headerInterceptor != null),
        _client = client {
    _client?.interceptors?.add(loggingInterceptor);
    _client?.interceptors?.add(headerInterceptor);
  }

  @override
  Future<M4eUser> getM4eUser(M4ePhoneNumber phoneNumber) async {
    try {
      // final _response = await _client.get('/users/${userId.value.toString()}');
      final _response = await _client
          .get('/users/${phoneNumber.country}/${phoneNumber.number}');

      return M4eUser.fromJson(_response.data);
    } on DioError catch (e) {
      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw M4eExceptionMessages.kConnectionTimeOutException;
      }

      if (e.response?.statusCode == 404) {
        throw M4eExceptionMessages.kNotFoundException;
      }

      if (e.response?.statusCode == 403) {
        throw M4eExceptionMessages.kUnauthorized;
      }

      if (e.response?.statusCode == 503) {
        throw M4eExceptionMessages.kServiceUnavailableException;
      }

      if (e.response?.statusCode == 500) {
        throw M4eExceptionMessages.kUndefinedException;
      }

      if (e.type == DioErrorType.RESPONSE) {
        throw M4eServerException.fromJson(e.response.data);
      }

      throw M4eServerException.fromString(e.message.toString());
    } on SocketException {
      throw M4eExceptionMessages.kNoInternetConnectionException;
    } on HttpException {
      throw M4eExceptionMessages.kHttpException;
    } on FormatException {
      throw M4eExceptionMessages.kHttpFormatException;
    } catch (e) {
      throw M4eServerException.fromString(e.toString());
    }
  }

  @override
  Future<M4eUniqueId> createM4eUser(
      M4eUniqueId userId, M4eUserForm m4eUserForm) async {
    assert(m4eUserForm != null, 'm4eUserForm should not be null');
    try {
      final _response = await _client.put(
        '/users/register/${userId.toString()}',
        data: json.encode(m4eUserForm.toJson()),
      );

      return M4eUniqueId.fromJson(_response.data);
    } on DioError catch (e) {
      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw M4eExceptionMessages.kConnectionTimeOutException;
      }

      if (e.response?.statusCode == 403) {
        throw M4eExceptionMessages.kUnauthorized;
      }

      if (e.response?.statusCode == 503) {
        throw M4eExceptionMessages.kServiceUnavailableException;
      }

      if (e.response?.statusCode == 500) {
        throw M4eExceptionMessages.kUndefinedException;
      }

      if (e.type == DioErrorType.RESPONSE) {
        throw M4eServerException.fromJson(e.response.data);
      }

      throw M4eServerException.fromString(e.message.toString());
    } on SocketException {
      throw M4eExceptionMessages.kNoInternetConnectionException;
    } on HttpException {
      throw M4eExceptionMessages.kHttpException;
    } on FormatException {
      throw M4eExceptionMessages.kHttpFormatException;
    } catch (e) {
      throw M4eServerException.fromString(e.toString());
    }
  }
}
