import 'dart:convert';
import 'dart:io';

import 'package:dio/dio.dart';
import 'package:flutter/foundation.dart';

import '../../core/core.dart';
import '../../models/models.dart';
import '../../m4e_forms/m4e_forms.dart';

/// Accessing the instantiated `WalletApi` methods
/// is done by passing its type to the GetIt instance
///
/// Example use of `WalletApi`:
/// ```
/// import 'package:m4edart/m4e_sdk_initialize.dart' as sdk;
///
/// sdk.m4e<WalletApi>().fetchUserWallets(userId);
/// ```
///
abstract class WalletApi {
  /// creates a search query with passed parameters
  Future<M4eUniqueId> queryUserWallets(
      SearchFormTypeCollection searchFormTypeCollection);

  /// Returns list of wallets for
  /// m4e user with passed-in entity wallet permissions
  Future<M4eWalletList> runSearchQuery(M4eUniqueId queryUniqueId);

  /// Returns wallet balance for specified wallet-id
  Future<M4eBalance> getWalletBalance(M4eUniqueId walletId);

  /// Creates a new wallet definition
  Future<M4eUniqueId> createWallet(
      M4eUniqueId userId, M4eWalletForm m4eWalletForm);
}

class WalletApiImpl implements WalletApi {
  final Dio _client;

  WalletApiImpl({
    @required Dio client,
    @required M4eLoggingInterceptors loggingInterceptor,
    @required M4eHeaderInterceptors headerInterceptor,
  })  : assert(client != null),
        assert(loggingInterceptor != null),
        assert(headerInterceptor != null),
        _client = client {
    // logs and appends authorization headers to requests
    _client.interceptors.add(loggingInterceptor);
    _client.interceptors.add(headerInterceptor);
  }

  Map<String, dynamic> _mapSearchFormTypeCollectionToActualType(
      SearchFormTypeCollection searchFormTypeCollection) {
    if (searchFormTypeCollection is SingleClauseSearchForm) {
      return searchFormTypeCollection.toJson();
    } else if (searchFormTypeCollection is MultipleClauseSearchForm) {
      return (searchFormTypeCollection as MultipleClauseSearchForm).toJson();
    } else {
      return (searchFormTypeCollection as MobileMoneyProviderSearchForm)
          .toJson();
    }
  }

  /// creates a search query with [SearchFormTypeCollection]
  Future<M4eUniqueId> queryUserWallets(
      SearchFormTypeCollection searchFormTypeCollection) async {
    try {
      final _response = await _client.put(
        '/search',
        data: json.encode(
            _mapSearchFormTypeCollectionToActualType(searchFormTypeCollection)),
      );

      return M4eUniqueId.fromJson(_response.data);
    } on DioError catch (e) {
      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw M4eExceptionMessages.kConnectionTimeOutException;
      }

      if (e.response?.statusCode == 403) {
        throw M4eExceptionMessages.kUnauthorized;
      }

      if (e.response?.statusCode == 503) {
        throw M4eExceptionMessages.kServiceUnavailableException;
      }

      if (e.response?.statusCode == 500) {
        throw M4eExceptionMessages.kUndefinedException;
      }

      if (e.type == DioErrorType.RESPONSE) {
        throw M4eServerException.fromJson(e.response.data);
      }

      throw M4eServerException.fromString(e.message.toString());
    } on SocketException {
      throw M4eExceptionMessages.kNoInternetConnectionException;
    } on HttpException {
      throw M4eExceptionMessages.kHttpException;
    } on FormatException {
      throw M4eExceptionMessages.kHttpFormatException;
    } catch (e) {
      throw M4eServerException.fromString(e.toString());
    }
  }

  /// Returns list of wallets for current authenticated
  /// m4e user with `OWNER` role entity-wallet attributes
  Future<M4eWalletList> runSearchQuery(M4eUniqueId queryUniqueId) async {
    try {
      final _response = await _client.get(
          '/search/run?id=${queryUniqueId.toString()}&epoch=${DateTime.now().toTimestampSinceEpoch()}');

      return M4eWalletList.fromJson(_response.data);
    } on DioError catch (e) {
      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw M4eExceptionMessages.kConnectionTimeOutException;
      }

      if (e.response?.statusCode == 403) {
        throw M4eExceptionMessages.kUnauthorized;
      }

      if (e.response?.statusCode == 503) {
        throw M4eExceptionMessages.kServiceUnavailableException;
      }

      if (e.response?.statusCode == 500) {
        throw M4eExceptionMessages.kUndefinedException;
      }

      if (e.type == DioErrorType.RESPONSE) {
        throw M4eServerException.fromJson(e.response.data);
      }

      throw M4eServerException.fromString(e.message.toString());
    } on SocketException {
      throw M4eExceptionMessages.kNoInternetConnectionException;
    } on HttpException {
      throw M4eExceptionMessages.kHttpException;
    } on FormatException {
      throw M4eExceptionMessages.kHttpFormatException;
    } catch (e) {
      throw M4eServerException.fromString(e.toString());
    }
  }

  @override
  Future<M4eBalance> getWalletBalance(M4eUniqueId walletId) async {
    assert(walletId != null, 'walletId should not be null');
    try {
      final _response = await _client.get(
          '/wallets/${walletId.toString()}/balance/${DateTime.now().toTimestampSinceEpoch()}');

      return M4eBalance.fromJson(_response.data);
    } on DioError catch (e) {
      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw M4eExceptionMessages.kConnectionTimeOutException;
      }

      if (e.response?.statusCode == 403) {
        throw M4eExceptionMessages.kUnauthorized;
      }

      if (e.response?.statusCode == 404) {
        throw M4eExceptionMessages.kNotFoundException;
      }

      if (e.response?.statusCode == 503) {
        throw M4eExceptionMessages.kServiceUnavailableException;
      }

      if (e.response?.statusCode == 500) {
        throw M4eExceptionMessages.kUndefinedException;
      }

      if (e.type == DioErrorType.RESPONSE) {
        throw M4eServerException.fromJson(e.response.data);
      }

      throw M4eServerException.fromString(e.message.toString());
    } on SocketException {
      throw M4eExceptionMessages.kNoInternetConnectionException;
    } on HttpException {
      throw M4eExceptionMessages.kHttpException;
    } on FormatException {
      throw M4eExceptionMessages.kHttpFormatException;
    } catch (e) {
      throw M4eServerException.fromString(e.toString());
    }
  }

  @override
  Future<M4eUniqueId> createWallet(
      M4eUniqueId userId, M4eWalletForm m4eWalletForm) async {
    assert(m4eWalletForm != null, 'm4eWalletForm should not be null');
    if (m4eWalletForm.type == M4eWalletType.M4E &&
        m4eWalletForm.detail is MobileMoneyWalletDetail) {
      throw M4eServerException.fromString(
          'M4eWalletForm [detail] should not be of type [MobileMoneyWalletDetail] when [M4eWalletType] is BASCULE! \nUse [M4eWalletDetail] as [M4eWalletDetailType] type instead!');
    }
    if (m4eWalletForm.type == M4eWalletType.M4E &&
        m4eWalletForm.id !=
            m4eWalletForm.detail.maybeMap(
                m4eWalletDetail: (detail) => detail.id, orElse: () => null)) {
      throw M4eServerException.fromString(
          'M4eWalletForm [id] and [detail][m4eWalletDetail] `id` should should be of the same value!');
    }

    try {
      final _response = await _client.put(
        '/wallets/${userId.toString()}/${DateTime.now().toTimestampSinceEpoch()}',
        data: json.encode(m4eWalletForm.toJson()),
      );

      return M4eUniqueId.fromJson(_response.data);
    } on DioError catch (e) {
      if (e.type == DioErrorType.CONNECT_TIMEOUT) {
        throw M4eExceptionMessages.kConnectionTimeOutException;
      }

      if (e.response?.statusCode == 403) {
        throw M4eExceptionMessages.kUnauthorized;
      }

      if (e.response?.statusCode == 503) {
        throw M4eExceptionMessages.kServiceUnavailableException;
      }

      if (e.response?.statusCode == 500) {
        throw M4eExceptionMessages.kUndefinedException;
      }

      if (e.type == DioErrorType.RESPONSE) {
        throw M4eServerException.fromJson(e.response.data);
      }

      throw M4eServerException.fromString(e.message.toString());
    } on SocketException {
      throw M4eExceptionMessages.kNoInternetConnectionException;
    } on HttpException {
      throw M4eExceptionMessages.kHttpException;
    } on FormatException {
      throw M4eExceptionMessages.kHttpFormatException;
    } catch (e) {
      throw M4eServerException.fromString(e.toString());
    }
  }
}
